import toml
import argparse
import numpy as np
import rdpolicy as rp
from mectools.optim import anneal_parallel

parser = argparse.ArgumentParser(description='Find optimal policies using parallel annealing.')
parser.add_argument('--target', type=str, default=None, help='data moments to target')
parser.add_argument('--params', type=str, default=None, help='parameter values to use')
parser.add_argument('--output', type=str, default=None, help='where to output estimates')
parser.add_argument('--weights', type=str, default=None, help='weight values to use')
parser.add_argument('--scale', type=float, default=0.05, help='initial scale for optimization')
parser.add_argument('--tol', type=float, default=1e-3, help='convergence tolerance for optimization')
parser.add_argument('--wait', type=int, default=100, help='wait time for convergence step')
parser.add_argument('--seed', type=int, default=None, help='random seed to use for annealing')
parser.add_argument('--threads', type=int, default=5, help='number of parallel threads to run')
parser.add_argument('--dryrun', action='store_true', help='run on dummy quadratic objective')
args = parser.parse_args()

# solver options
opts = {
    'tol'   : 1e-12,
    'output': False,
    'bail'  : True,
    'reps'  : 1500,
}

# objective function to maximize
def objective(par_var, par_fix, mmt_dat, wgt_dat):
    params = {**par_var, **par_fix}

    # bounds checking
    if np.any([(v < 0) for v in params.values()]) or params['p'] >= 1:
        print('Invalid parameters')
        return np.inf

    # calculate equilibrium and simulate moments
    econ = rp.Economy(params=params)
    eq, _ = econ.equilibrium(**opts)
    panel = econ.generate_panel(eq)
    mmt_sim = econ.calc_moments(panel, infer=True, nthread=1)

    # objective value
    mmt_sim_vec = np.array([mmt_sim[k] for k in mmt_dat])
    mmt_dat_vec = np.array([mmt_dat[k] for k in mmt_dat])
    mmt_wgt_vec = np.array([wgt_dat[k] for k in mmt_dat])
    mmt_obj = np.sum(mmt_wgt_vec*(mmt_sim_vec-mmt_dat_vec)**2)

    return mmt_obj

# load inital values
par_init = rp.file_or_dict(args.params if args.params is not None else rp.params_default_fname)
mmt_data = rp.file_or_dict(args.target if args.target is not None else rp.target_default_fname)
wgt_data = rp.file_or_dict(args.weights) if args.weights is not None else {k: 1/v**2 for k, v in mmt_data.items()}

# find variable params
par_fix = {k: v for k, v in par_init.items() if k not in rp.var_params}
par_var = {k: v for k, v in par_init.items() if k in rp.var_params}

# various args
targs = {
    'scale': args.scale,
    'rand': 'mul',
    'tol': args.tol,
    'wait': args.wait,
    'seed': args.seed,
}

# set up objective
if args.dryrun:
    obj = lambda par: sum([(v-1)**2 for v in par.values()])
else:
    obj = lambda par: objective(par, par_fix, mmt_data, wgt_data)

# run optimizer
parf, objf = anneal_parallel(obj, par_var, N=args.threads, **targs)
parf.update(par_fix)

# output solution
if args.output is not None:
    with open(args.output, 'w+') as fout:
        toml.dump(parf, fout)
else:
    print(toml.dumps(parf))
