import argparse
import toml
import numpy as np
import rdpolicy as rp
from mectools.optim import anneal_parallel

parser = argparse.ArgumentParser(description='Find optimal policies using parallel annealing.')
parser.add_argument('--params', type=str, help='parameter values to use')
parser.add_argument('--ptype', type=str, help='policy type to work with')
parser.add_argument('--pfile', type=str, default='policy/default_policies.toml', help='policy type to work with')
parser.add_argument('--ptag', type=str, default='baseline', help='tag to identify params by')
parser.add_argument('--control', action='store_true', help='control production (rather than monopoly)')
parser.add_argument('--qmax', type=float, default=None, help='maximal value for qgrid')
parser.add_argument('--scale', type=float, default=0.1, help='initial scale for optimization')
parser.add_argument('--tol', type=float, default=1e-3, help='convergence tolerance for optimization')
parser.add_argument('--wait', type=int, default=50, help='wait time for convergence step')
parser.add_argument('--threads', type=int, default=5, help='number of parallel threads to run')
args = parser.parse_args()

# solver options
opts = {
    'tol'   : 1e-12,
    'output': False,
    'bail'  : True,
    'reps'  : 1500,
}

# objective function to maximize
def objective(pol, params, ptype, q_max=None, **kwargs):
    # set up solver
    params1 = params.copy()
    if q_max is not None:
        params1['q_max'] = q_max
    econ = rp.Economy(params=params1)
    eval_func = getattr(econ, f'equilibrium_{ptype}')

    # construct arguments
    args = {**pol, **opts, **kwargs}

    # evaluate objective
    try:
        eq, ts = eval_func(**args)
        welf = eq.welfare
        if eq.err > eq.tol or welf > 1e5:
            welf = -np.inf
    except Exception as e:
        welf = -np.inf
        raise(e)
    return welf

# initial guesses
params = rp.file_or_dict(args.params) if args.params is not None else {}
policy = rp.file_or_dict(args.pfile)[args.ptype]

# various args
targs = {'scale': args.scale, 'tol': args.tol, 'wait': args.wait}
oargs = {'control': args.control, 'q_max': args.qmax}

# generate tag
stag = 'control' if args.control else 'monopoly'
tag = f'{args.ptag}_{stag}'

# optimize welfare
obj = lambda pol: -objective(pol, params, args.ptype, **oargs)
polf, welf = anneal_parallel(obj, policy, N=args.threads, **targs)
polf['welfare'] = -welf

# print final results
print(toml.dumps({args.ptype: {tag: polf}}))
