# paper graphs

# import matplotlib for saving
import matplotlib as mpl
import matplotlib.style as style
mpl.use('agg')
style.use('algo/clean.mplstyle')
import matplotlib.pyplot as plt

# other imports
import os
import numpy as np
import pandas as pd
import rdpolicy as rp
from scipy.interpolate import PchipInterpolator

# options
tmax0 = 30
times0 = [2, 5, 15]
ext0 = ['pdf']
dpi0 = 300
figsize = 5, 4.5
graph_dir = 'images'
latex_dir = 'latex'

times0_s = ','.join([str(t) for t in times0])
ext0_s = ','.join(ext0)

##
## plotting
##

def smooth(x, y, N=200):
    x_min, x_max = x.min(), x.max()
    x1 = np.linspace(x_min, x_max, N)
    y1 = PchipInterpolator(x, y)(x1)
    return x1, y1

def ssmooth(s, **kwargs):
    x1, y1 = smooth(s.index.values, s.values, **kwargs)
    return pd.Series(y1, index=x1)

def smooth_plot(x, y, ax, N=200, **kwargs):
    x1, y1 = smooth(x, y, N=N)
    return ax.plot(x1, y1, **kwargs)

##
## wedges
##

def plot_wedge_time(econ, ts, tmax=tmax0):
    ts1 = ts.loc[2:tmax, ['tax', 'sub', 'subt']].apply(ssmooth, axis=0)

    fig1, ax1 = plt.subplots(figsize=figsize)
    ts1['tax'].plot(ax=ax1)

    fig2, ax2 = plt.subplots(figsize=figsize)
    ts1[['sub', 'subt']].rename({
        'sub': 'Net',
        'subt': 'Gross'
    }, axis=1).plot(ax=ax2)

    # ax1.set_title('Tax Wedge')
    ax1.set_xlabel('Firm Age')
    # ax1.set_ylim(bottom=0)

    # ax2.set_title('Subsidy Wedge')
    ax2.set_xlabel('Firm Age')
    # ax2.set_ylim(bottom=0)

    fig1.subplots_adjust(bottom=0.20)
    fig2.subplots_adjust(bottom=0.20)

    return fig1, fig2

def plot_wedge_type(econ, eq, observe=False, times=times0):
    fig1, ax1 = plt.subplots(figsize=figsize)
    fig2, ax2 = plt.subplots(figsize=figsize)
    ax1.margins(x=0.0, y=0.2)
    ax2.margins(x=0.0, y=0.2)

    theta1_cond = lambda x: np.sum(econ.th11_dist*x, axis=1)/np.sum(econ.th11_dist, axis=1)
    theta2_cond = lambda x: np.sum(econ.th12_dist*x, axis=(1, 2))/np.sum(econ.th12_dist, axis=(1, 2))
    thetaq_cond = lambda x: np.sum(eq.dist*x, axis=(1, 2, 3))/np.sum(eq.dist, axis=(1, 2, 3))

    tax_cond = theta2_cond(eq.tax)
    subt_cond = theta1_cond(eq.subt)
    sub_cond = theta1_cond(eq.sub)

    if observe:
        s_xvar = theta1_cond(eq.rc)
        t_xvar = thetaq_cond(eq.pi)
        s_xlab = 'R&D Investment'
        t_xlab = 'Firm Profit'
    else:
        s_xvar = np.tile(econ.ltheta_vec, (econ.T, 1))
        t_xvar = np.tile(econ.ltheta_vec, (econ.T, 1))
        s_xlab = 'Firm Type $\\log(\\theta)$'
        t_xlab = 'Firm Type $\\log(\\theta)$'

    for t in times:
        label = f't = {t}'
        smooth_plot(t_xvar[t-1,:], tax_cond[t-1,:], ax1, label=label)
        line, = smooth_plot(s_xvar[t-1,:], subt_cond[t-1,:], ax2, label=label)
        smooth_plot(s_xvar[t-1,:], sub_cond[t-1,:], ax2, label=label, linestyle='--', color=line.get_color())

    # phantom lines for legend
    line_gross, = ax2.plot([], [], color='black')
    line_net, = ax2.plot([], [], color='black', linestyle='--')

    ax1.legend(loc='best')
    # ax1.set_title('Tax Wedge')
    ax1.set_xlabel(t_xlab)
    # ax1.set_ylim(bottom=0)

    ax2.legend([line_net, line_gross], ['Net', 'Gross'], loc='best')
    # ax2.set_title('Net Subsidy Wedge')
    ax2.set_xlabel(s_xlab)
    # ax2.set_ylim(bottom=0)

    fig1.subplots_adjust(bottom=0.20)
    fig2.subplots_adjust(bottom=0.20)

    return fig1, fig2

##
## allocations
##

def plot_alloc_time(econ, ts, tmax=tmax0):
    ts1 = ts.loc[2:tmax, ['r', 'l', 'lam', 'pi']].apply(ssmooth, axis=0)

    fig1, ax1 = plt.subplots(figsize=figsize)
    ts1['r'].rename('R&D effort ($r$)').plot(ax=ax1)
    ts1['l'].rename('Manager effort ($\\ell$)').plot(ax=ax1)

    fig2, ax2 = plt.subplots(figsize=figsize)
    ts1['lam'].rename('Step size ($\\lambda$)').plot(ax=ax2)
    ts1['pi'].rename('Profit ($\\pi$)').plot(ax=ax2)

    ax1.legend(loc='best')
    # ax1.set_title('Inputs')
    ax1.set_xlabel('Firm Age')
    ax1.set_ylim(bottom=0)

    ax2.legend(loc='best')
    # ax2.set_title('Outputs')
    ax2.set_xlabel('Firm Age')
    ax2.set_ylim(bottom=0)

    fig1.subplots_adjust(bottom=0.20)
    fig2.subplots_adjust(bottom=0.20)

    return fig1, fig2

def plot_alloc_type(econ, eq, times=times0):
    fig1, ax1 = plt.subplots(figsize=figsize)
    fig2, ax2 = plt.subplots(figsize=figsize)

    ax1.margins(x=0.0, y=0.2)
    ax2.margins(x=0.0, y=0.2)

    theta1_cond = lambda x: np.sum(econ.th11_dist*x, axis=1)/np.sum(econ.th11_dist, axis=1)
    theta2_cond = lambda x: np.sum(econ.th12_dist*x, axis=(1, 2))/np.sum(econ.th12_dist, axis=(1, 2))

    l_cond = theta2_cond(eq['l'])
    r_cond = theta1_cond(eq['r'])

    for t in times:
        label = f't = {t}'
        smooth_plot(econ.ltheta_vec, l_cond[t-1,:], ax1, label=label)
        smooth_plot(econ.ltheta_vec, r_cond[t-1,:], ax2, label=label)

    ax1.legend(loc='best')
    # ax1.set_title('Average $\\ell$')
    ax1.set_xlabel('Firm Type $\\log(\\theta)$')
    ax1.set_ylim(bottom=0)

    # ax2.set_title('Average $r$')
    ax2.set_xlabel('Firm Type $\\log(\\theta)$')
    ax2.set_ylim(bottom=0)

    fig1.subplots_adjust(bottom=0.20)
    fig2.subplots_adjust(bottom=0.20)

    return fig1, fig2

##
## marginals
##

def plot_marginal_time(econ, ts, tmax=tmax0):
    ts1 = ts.loc[2:tmax]

    fig1, ax1 = plt.subplots(figsize=figsize)
    fig2, ax2 = plt.subplots(figsize=figsize)
    ts1['mt'].plot(ax=ax1)
    ts1['ms'].plot(ax=ax2)

    # ax1.set_title('Marginal Tax Rate')
    ax1.set_xlabel('Firm Age')
    # ax1.set_ylim(bottom=0)

    # ax2.set_title('Marginal Subsidy Rate')
    ax2.set_xlabel('Firm Age')
    # ax2.set_ylim(bottom=0)

    fig1.subplots_adjust(bottom=0.20)
    fig2.subplots_adjust(bottom=0.20)

    return fig1, fig2

def plot_marginal_type(econ, eq, observe=False, times=times0):
    fig1, ax1 = plt.subplots(figsize=figsize)
    fig2, ax2 = plt.subplots(figsize=figsize)
    ax1.margins(x=0.0, y=0.2)
    ax2.margins(x=0.0, y=0.2)

    theta1_cond = lambda x: np.sum(econ.th11_dist*x, axis=1)/np.sum(econ.th11_dist, axis=1)
    theta2_cond = lambda x: np.sum(econ.th12_dist*x, axis=(1, 2))/np.sum(econ.th12_dist, axis=(1, 2))
    thetaq_cond = lambda x: np.sum(eq['dist']*x, axis=(1, 2, 3))/np.sum(eq['dist'], axis=(1, 2, 3))

    mt_cond = theta2_cond(eq['mt_new'])
    ms_cond = theta1_cond(eq['ms'])

    if observe:
        s_xvar = theta1_cond(eq['rc'])
        t_xvar = thetaq_cond(eq['pi'])
        s_xlab = 'R&D Investment'
        t_xlab = 'Firm Profit'
    else:
        s_xvar = np.tile(econ.ltheta_vec, (econ.T, 1))
        t_xvar = np.tile(econ.ltheta_vec, (econ.T, 1))
        s_xlab = 'Firm Type $\\log(\\theta)$'
        t_xlab = 'Firm Type $\\log(\\theta)$'

    for t in times:
        label = f't = {t}'
        ax1.plot(t_xvar[t,:], mt_cond[t,:], label=label)
        ax2.plot(s_xvar[t,:], ms_cond[t,:], label=label)

    ax1.legend(loc='best')
    # ax1.set_title('Marginal Tax Rate')
    ax1.set_xlabel(t_xlab)
    # ax1.set_ylim(bottom=0)

    # ax2.set_title('Marginal Subsidy Rate')
    ax2.set_xlabel(s_xlab)
    # ax2.set_ylim(bottom=0)

    fig1.subplots_adjust(bottom=0.20)
    fig2.subplots_adjust(bottom=0.20)

    return fig1, fig2

##
## figure generation
##

def slide_graphs(econ, eq, ts, prefix, tmax=tmax0, times=times0, scale=False, ext=ext0, dpi=dpi0):
    # figure A

    fig1, fig2 = plot_wedge_time(econ, ts, tmax=tmax)
    fig1.savefig(f'{graph_dir}/{prefix}_wedge_tax_time.{ext}', dpi=dpi)
    fig2.savefig(f'{graph_dir}/{prefix}_wedge_sub_time.{ext}', dpi=dpi)

    fig1, fig2 = plot_wedge_type(econ, eq, observe=True, times=times)
    fig1.savefig(f'{graph_dir}/{prefix}_wedge_tax_obsv.{ext}', dpi=dpi)
    fig2.savefig(f'{graph_dir}/{prefix}_wedge_sub_obsv.{ext}', dpi=dpi)

    # figure B

    fig1, fig2 = plot_alloc_time(econ, ts, tmax=tmax)
    fig1.savefig(f'{graph_dir}/{prefix}_alloc_choice_time.{ext}', dpi=dpi)
    fig2.savefig(f'{graph_dir}/{prefix}_alloc_result_time.{ext}', dpi=dpi)

    fig1, fig2 = plot_alloc_type(econ, eq, times=times)
    fig1.savefig(f'{graph_dir}/{prefix}_alloc_ell_type.{ext}', dpi=dpi)
    fig2.savefig(f'{graph_dir}/{prefix}_alloc_rnd_type.{ext}', dpi=dpi)

    # internal graphs

    fig1, fig2 = plot_wedge_type(econ, eq, observe=False, times=times)
    fig1.savefig(f'{graph_dir}/{prefix}_wedge_tax_type.{ext}', dpi=dpi)
    fig2.savefig(f'{graph_dir}/{prefix}_wedge_sub_type.{ext}', dpi=dpi)

    # marginals

    # fig1, fig2 = plot_marginal_time(econ, ts, tmax=tmax)
    # fig1.savefig(f'{graph_dir}/{prefix}_marginal_tax_time.{ext}', dpi=dpi)
    # fig2.savefig(f'{graph_dir}/{prefix}_marginal_sub_time.{ext}', dpi=dpi)

    # fig1, fig2 = plot_marginal_type(econ, eq, observe=True, times=times)
    # fig1.savefig(f'{graph_dir}/{prefix}_marginal_tax_obsv.{ext}', dpi=dpi)
    # fig2.savefig(f'{graph_dir}/{prefix}_marginal_sub_obsv.{ext}', dpi=dpi)

    # fig1, fig2 = plot_marginal_type(econ, eq, observe=False, times=times)
    # fig1.savefig(f'{graph_dir}/{prefix}_marginal_tax_type.{ext}', dpi=dpi)
    # fig2.savefig(f'{graph_dir}/{prefix}_marginal_sub_type.{ext}', dpi=dpi)

    plt.close('all')

def gen_graphs(par, prefix, control=False, psub_wedge=False, ext=ext0, **kwargs):
    econ = rp.Economy(params=par)
    eq, ts = econ.optimal(control=control, psub_wedge=psub_wedge)
    for e in ext:
        slide_graphs(econ, eq, ts, prefix, ext=e, **kwargs)

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Generate tables for back to basics.')
    parser.add_argument('--params', type=str, default=rp.params_default_fname, help='parameter set to use')
    parser.add_argument('--prefix', type=str, default='test', help='prefix for output graphs')
    parser.add_argument('--times', type=str, default=times0_s, help='times to plot for wedges')
    parser.add_argument('--ext', type=str, default=ext0_s, help='extension for image output')
    parser.add_argument('--dpi', type=int, default=dpi0, help='dpi for image output')
    parser.add_argument('--control', action='store_true', help='can we control quantity for wedges')
    parser.add_argument('--psub-wedge', action='store_true', help='wedges relative to price subsidy case')
    parser.add_argument('--tmax', type=int, default=tmax0, help='time horizon to display')
    args = parser.parse_args()

    times = [int(t) for t in args.times.split(',')]
    ext = args.ext.split(',')

    print(f'params = {args.params}')
    print(f'prefix = {args.prefix}')
    print(f'control = {args.control}')
    print(f'psub_wedge = {args.psub_wedge}')
    print(f'tmax = {args.tmax}')
    print(f'times = {times}')
    print(f'ext = {ext}')

    if not os.path.exists(graph_dir):
        os.mkdir(graph_dir)

    if not os.path.exists(latex_dir):
        os.mkdir(latex_dir)

    gen_graphs(
        args.params, args.prefix, control=args.control,
        psub_wedge=args.psub_wedge, tmax=args.tmax, times=times, ext=ext
    )

    with open('algo/graphs.tex') as fin:
        figure_template = fin.read()
    latex_out = figure_template % {'prefix': args.prefix}

    with open(f'{latex_dir}/{args.prefix}.tex', 'w+') as fout:
        fout.write(latex_out)
