import os
import toml
import numpy as np
import pandas as pd
import rdpolicy as rp

pd.options.display.max_colwidth = 1000

fmt = '%.3f'

latex_template = """\\documentclass[10pt]{article}

\\usepackage{amsmath}
\\usepackage{amssymb}
\\usepackage{booktabs}
\\usepackage[left=1in,right=1in,top=1in,bottom=1in]{geometry}

\\setlength{\\parindent}{0cm}
\\setlength{\\parskip}{0.2cm}

\\begin{document}

%s

\\end{document}"""

par_names = {
    'R': '$R$',
    'alpha': '$\\alpha$',
    'beta': '$\\beta$',
    'delta': '$\\delta$',
    'eps_std': '$\\sigma_{\\epsilon}$',
    'eta': '$\\eta$',
    'gamma': '$\\gamma$',
    'kappal': '$\\kappa_{l}$',
    'kappar': '$\\kappa_{r}$',
    'ltheta1_range': '$\\Theta^1$',
    'ltheta_mu': '$\\mu_{\\theta}$',
    'p': '$p$',
    'qbar0': '$q_0$',
    'r0': '$r_0$',
    'rho': '$\\rho_{\\theta r}$',
    'zeta': '$\\zeta$'
}

par_desc = {
    'R': 'Interest rate',
    'alpha': 'R\&D share',
    'beta': 'Knowledge share',
    'delta': 'Intangibles depreciation',
    'eps_std': 'Type variance',
    'eta': 'R\&D cost elasticity',
    'gamma': 'Effort cost elasticity',
    'kappal': 'Scale of disutility',
    'kappar': 'Scale of R\&D cost',
    'ltheta1_range': 'Support for $\\theta_1$',
    'ltheta_mu': 'Leve of types',
    'p': 'Type persistence',
    'qbar0': 'Initial intangibles',
    'r0': 'Initial R\&D stock',
    'rho': 'R\&D-type substitution',
    'zeta': 'Production externality'
}

mmt_names = {
    'rnd_elasticity': 'M1. Patent quality-R\&D elasticity',
    'rndi_mean': 'M2. R\&D/Sales median',
    'sales_growth': 'M3. Sales growth (DHS) mean',
    'patqual_within_cov': 'M4. Within-firm patent quality coeff of var',
    'patqual_across_yng_cov': 'M5. Across-firm patent quality coeff of var (young)',
    'patqual_across_old_cov': 'M6. Across-firm patent quality coeff of var (old)',
    'patqual_ratio': 'M7. Patent quality young/old',
    'spillover_reg': 'M8. Spillover regression coefficient',
    'subsidies_reg': 'M9. Subsidy regression coefficient'
}

def param_table(params):
    par = toml.load(open(params))
    df = pd.DataFrame({'Value': par})
    df['Symbol'] = pd.Series(par_names)
    df['Parameter'] = pd.Series(par_desc)
    return df[['Parameter', 'Symbol', 'Value']].to_latex(escape=False, float_format=fmt, index=False)

def moment_table(params, moments):
    df = rp.get_moments(params, target=moments)
    df = df.rename(columns={'data': 'Target', 'simul': 'Simulation'})
    df['Moment'] = pd.Series(mmt_names)
    return df[['Moment', 'Target', 'Simulation']].to_latex(escape=False, float_format=fmt, index=False)

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Generate tables for back to basics.')
    parser.add_argument('--output', type=str, default='latex/tables.tex', help='output file for latex')
    parser.add_argument('--params', type=str, default=rp.params_default_fname, help='parameter set to use')
    parser.add_argument('--moments', type=str, default=rp.target_default_fname, help='moment targets to use')
    args = parser.parse_args()

    par = param_table(args.params)
    mmt = moment_table(args.params, args.moments)

    latex = ''
    latex += par + '\n'
    latex += mmt + '\n'

    with open(args.output, 'w+') as outp:
        outp.write(latex_template % latex)
