# R&D policy - AHS

import toml
import numpy as np
import pandas as pd
import scipy.stats as stats
import statsmodels.formula.api as smf
import mectools.endy as nd
import mectools.bundle as bn
from collections import OrderedDict
from multiprocessing import Pool

# kill off warnings
import warnings
warnings.filterwarnings('ignore')

# why must they
np.concat = np.concatenate
def log(x):
    y = np.log(x)
    y[np.isinf(y)] = np.nan
    return y

##
## solver defaults
##

reps0 = 2000 # default number of reps
tol0 = 1e-12 # default convergence tolerance

##
## simulation defaults
##

SEED = 2537801916 # seed for simulations
NF0 = 32*1024 # number of firms in simulation

##
## estimation config
##

age_cut = 5 # age cutoff for young/old
sim_cut = 50 # how many periods to use in estimation

var_params = [
    'rho', 'alpha', 'kappar', 'kappal', 'p', 'eps_std', 'ltheta1_range',
    'zeta', 'gamma'
]
mmt_names = [
    'rnd_elasticity', 'rndi_mean', 'sales_growth', 'patqual_within_cov',
    'patqual_across_yng_cov', 'patqual_across_old_cov', 'patqual_ratio',
    'spillover_reg', 'subsidies_reg'
]

##
## file names
##

params_default_fname = 'params/estimated_allfirms.toml'
target_default_fname = 'data/moments_allfirms.toml'

params_default = {
    'T': 100, # time horizon
    'Q': 16, # q grid size
    'q_min': 0.0, # minimum q grid value
    'q_max': 500.0, # maximum q grid value
    'N': 8, # theta grid size
    'N1': 8, # theta1 grid size
    'dampen': 0.5, # dampening on qbar update
    'pct_max': 0.2, # max update percent
    'val_min': 0, # minimum numeric value
    'val_max': 1000, # maximum numeric value
    'theta_range0': 2.0, # theta range scale
    'qboost': 1.0 # externality bonus
}

##
## model parameters
##

# alpha: type/research share
# rho: type/research elasticity (1.2)
# gamma: elasticity of managerial cost
# eps_std: type shock variance
# p: type persistence
# R: interest rate
# kappar: R&D cost scale
# kappal: Labor cost scale
# eta: R&D cost elasticity (3.0)
# beta: production shares (0.65)
# zeta: spillover exponent (in cost)
# delta: quality depreciation
# ltheta_mu: mean reversion point for theta (and mean of theta1)
# ltheta1_range: log span of theta1 (uniform) distribution
# qbar0: initial average productivity level
# r0: initial research stock

##
## policy
##

tau_base = 0.23
sub_base = 0.19

dtau_base = lambda pi, rc: tau_base*np.ones_like(pi)
ttau_base = lambda pi, rc: tau_base*pi
dsub_base = lambda rc: sub_base*np.ones_like(rc)
tsub_base = lambda rc: sub_base*rc
dcross_base = lambda pi, rc: np.zeros_like(pi)

##
## useful functions
##

def load_toml(fname):
    with open(fname) as fid:
        d = toml.load(fid, _dict=OrderedDict)
    return d

def save_toml(d, fname):
    with open(fname, 'w') as fid:
        toml.dump(d, fid)

def file_or_dict(fod):
    return load_toml(fod) if type(fod) is str else fod

def stack_dicts(dicts, ref_idx=0, agg_func=np.array, reverse=False):
    ref_dict = dicts[ref_idx]
    keys = ref_dict.keys()
    defaults = {k: np.nan*np.zeros_like(v) for (k, v) in ref_dict.items()}
    ndicts = len(dicts)
    idxs = range(ndicts)
    if reverse:
        idxs = reverse(idxs)
    return {k: agg_func([dicts[i].get(k, defaults[k]) for i in idxs]) for k in keys}

##
## parallel operators
##

def regression_op(params, NF, seed):
    t_vec = np.arange(1, sim_cut+1)
    t_rev = np.arange(sim_cut, 0, -1)
    f_vec = np.arange(NF)

    deprec = 0.85
    accum = lambda s, v0=0: (np.r_[v0,s[:-1]]*(deprec**(t_rev-1))).cumsum()/(deprec**(t_rev-1))

    fpanel = []
    base_cols = ['age', 'firm', 'qval', 'rnd', 'lam']

    econ = Economy(params=params)
    eq, ts = econ.equilibrium(output=False)
    panel = econ.generate_panel(eq, NF=NF, seed=seed)

    panel['age'] = t_vec[:,None].repeat(NF, axis=1)
    panel['firm'] = f_vec[None,:].repeat(sim_cut, axis=0)
    dpanel = pd.DataFrame({k: panel[k].flatten() for k in panel if k in base_cols})

    qbser = pd.Series(ts['qbar'])
    dpanel = dpanel.join(qbser, on='age')
    dpanel['kappar'] = econ.kappar

    dpanel['age1'] = dpanel['age'] - 1
    merge_lag = dpanel.set_index(['firm', 'age'])[['rnd', 'lam']]
    dpanel = dpanel.join(merge_lag, on=['firm', 'age1'], rsuffix='_lag')
    dpanel = dpanel.drop('age1', axis=1)

    dpanel['rnd_cum'] = dpanel.groupby('firm')['rnd'].transform(lambda s: accum(s, econ.r0))
    dpanel['lam_cum'] = dpanel.groupby('firm')['lam'].transform(lambda s: accum(s, 1))
    dpanel['sales'] = (econ.psi/econ.beta)*dpanel['qbar']**(econ.zeta*(1.0-econ.beta)/econ.beta)*dpanel['qval']

    age_agg = dpanel.groupby('age')[['rnd', 'rnd_cum', 'lam', 'lam_cum']].mean().add_suffix('_avg')
    dpanel = dpanel.join(age_agg, on='age')

    return dpanel

##
## policy forms
##

def policy_lin(tau, sub):
    # core linear functions
    dtau = lambda pi, rc: tau*np.ones_like(pi)
    ttau = lambda pi, rc: tau*pi
    dsub = lambda rc: sub*np.ones_like(rc)
    tsub = lambda rc: sub*rc

    return dtau, ttau, dsub, tsub

def policy_linx(tau0, tau1, sub0):
    # rescaling
    tau1 /= 10

    trate = lambda rc: tau0 - tau1*rc
    taubnd = lambda rc: np.minimum(trate(rc), 1)
    dtaubnd = lambda rc: np.where(trate(rc) < 1, -tau1, 0)

    dtau = lambda pi, rc: taubnd(rc)*np.ones_like(pi)
    ttau = lambda pi, rc: taubnd(rc)*pi
    dsub = lambda rc: sub0*np.ones_like(rc)
    tsub = lambda rc: sub0*rc
    dcross = lambda pi, rc: dtaubnd(rc)*pi

    return dtau, ttau, dsub, tsub, dcross

def policy_hsv(tau0, tau1, sub0, sub1, tau2, sub2):
    # positivity transforms
    tau2 = np.exp(tau2)
    sub2 = np.exp(sub2)

    # core HSV+int functions
    dtau0 = lambda pi, rc: tau0 - tau1*pi**tau2
    ttau0 = lambda pi, rc: tau0*pi - (tau1/(1+tau2))*pi**(1+tau2)
    dsub0 = lambda rc: sub0 - sub1*rc**sub2
    tsub0 = lambda rc: sub0*rc - (sub1/(1+sub2))*rc**(1+sub2)

    # bound pi from below
    pibar = np.maximum(0, (tau0-1)/tau1)**(1/tau2)
    if tau1 >= 0:
        dtau = lambda pi, rc: np.where(pi < pibar, 1, dtau0(pi, rc))
        ttau = lambda pi, rc: np.where(pi < pibar, pi, pibar + (ttau0(pi, rc)-ttau0(pibar, rc)))
    else:
        dtau = lambda pi, rc: np.where(pi < pibar, dtau0(pi, rc), 1)
        ttau = lambda pi, rc: np.where(pi < pibar, ttau0(pi, rc), (pi-pibar) + ttau0(pibar, rc))

    # bound rc from below
    rcbar = np.maximum(0, (sub0-1)/sub1)**(1/sub2)
    if sub1 >= 0:
        dsub = lambda rc: np.where(rc < rcbar, 1, dsub0(rc))
        tsub = lambda rc: np.where(rc < rcbar, rc, rcbar + (tsub0(rc)-tsub0(rcbar)))
    else:
        dsub = lambda rc: np.where(rc < rcbar, dsub0(rc), 1)
        tsub = lambda rc: np.where(rc < rcbar, tsub0(rc), (rc-rcbar) + tsub0(rcbar))

    return dtau, ttau, dsub, tsub

def policy_hsvx(tau0, tau1, tau2, tau3, sub0, sub1, sub2):
    # positivity transforms
    tau2 = np.exp(tau2)
    sub2 = np.exp(sub2)

    # rescale interacation
    tau3 /= 10

    # core HSVX functions
    dtau0 = lambda pi, rc: (tau0-(tau3/(1+sub2))*rc**(1+sub2)) - tau1*pi**tau2
    ttau0 = lambda pi, rc: (tau0-(tau3/(1+sub2))*rc**(1+sub2))*pi - (tau1/(1+tau2))*pi**(1+tau2)
    dsub0 = lambda rc: sub0 - sub1*rc**sub2
    tsub0 = lambda rc: sub0*rc - (sub1/(1+sub2))*rc**(1+sub2)
    dcross0 = lambda pi, rc: -tau3*(rc**sub2)*pi

    # bound pi from above/below
    pibar = lambda rc: np.maximum(0, (tau0-(tau3/(1+sub2))*rc**(1+sub2)-1)/tau1)**(1/tau2)
    if tau1 >= 0:
        dtau = lambda pi, rc: np.where(pi < pibar(rc), 1, dtau0(pi, rc))
        ttau = lambda pi, rc: np.where(pi < pibar(rc), pi, pibar(rc) + (ttau0(pi, rc)-ttau0(pibar(rc), rc)))
        dcross = lambda pi, rc: np.where(pi < pibar(rc), 0, dcross0(pi, rc)-dcross0(pibar(rc), rc))
    else:
        dtau = lambda pi, rc: np.where(pi < pibar(rc), dtau0(pi, rc), 1)
        ttau = lambda pi, rc: np.where(pi < pibar(rc), ttau0(pi, rc), (pi-pibar(rc)) + ttau0(pibar(rc), rc))
        dcross = lambda pi, rc: np.where(pi < pibar(rc), dcross0(pi, rc)-dcross0(pibar(rc), rc), 0)

    # bound rc from above/below
    rcbar = np.maximum(0, (sub0-1)/sub1)**(1/sub2)
    if sub1 >= 0:
        dsub = lambda rc: np.where(rc < rcbar, 1, dsub0(rc))
        tsub = lambda rc: np.where(rc < rcbar, rc, rcbar + (tsub0(rc)-tsub0(rcbar)))
    else:
        dsub = lambda rc: np.where(rc < rcbar, dsub0(rc), 1)
        tsub = lambda rc: np.where(rc < rcbar, tsub0(rc), (rc-rcbar) + tsub0(rcbar))

    return dtau, ttau, dsub, tsub, dcross

def policy_linhsv(tau, sub0, sub1, sub2):
    # positivity transforms
    sub2 = np.exp(sub2)

    # core HSV+int functions
    dtau = lambda pi, rc: tau*np.ones_like(pi)
    ttau = lambda pi, rc: tau*pi
    dsub0 = lambda rc: sub0 - sub1*rc**sub2
    tsub0 = lambda rc: sub0*rc - (sub1/(1+sub2))*rc**(1+sub2)

    # bound rc from below
    rcbar = np.maximum(0, (sub0-1)/sub1)**(1/sub2)
    if sub1 >= 0:
        dsub = lambda rc: np.where(rc < rcbar, 1, dsub0(rc))
        tsub = lambda rc: np.where(rc < rcbar, rc, rcbar + (tsub0(rc)-tsub0(rcbar)))
    else:
        dsub = lambda rc: np.where(rc < rcbar, dsub0(rc), 1)
        tsub = lambda rc: np.where(rc < rcbar, tsub0(rc), (rc-rcbar) + tsub0(rcbar))

    return dtau, ttau, dsub, tsub

def policy_hsvlin(tau0, tau1, tau2, sub):
    # positivity transforms
    tau2 = np.exp(tau2)

    # core HSV+int functions
    dtau0 = lambda pi, rc: tau0 - tau1*pi**tau2
    ttau0 = lambda pi, rc: tau0*pi - (tau1/(1+tau2))*pi**(1+tau2)
    dsub = lambda rc: sub*np.ones_like(rc)
    tsub = lambda rc: sub*rc

    # bound pi from below
    pibar = np.maximum(0, (tau0-1)/tau1)**(1/tau2)
    if tau1 >= 0:
        dtau = lambda pi, rc: np.where(pi < pibar, 1, dtau0(pi, rc))
        ttau = lambda pi, rc: np.where(pi < pibar, pi, pibar + (ttau0(pi, rc)-ttau0(pibar, rc)))
    else:
        dtau = lambda pi, rc: np.where(pi < pibar, dtau0(pi, rc), 1)
        ttau = lambda pi, rc: np.where(pi < pibar, ttau0(pi, rc), (pi-pibar) + ttau0(pibar, rc))

    return dtau, ttau, dsub, tsub

##
## main interface class
##

class Economy:
    def __init__(self, params={}):
        # params can be updated
        self.set_params(params)

    # param handling
    def set_params(self, params={}, grid=True):
        self.params = params_default.copy()
        self.params.update(load_toml(params_default_fname))
        self.params.update(file_or_dict(params))
        for k, v in self.params.items():
            setattr(self, k, v)

        # derived stats
        self.psi = self.beta*(1.0-self.beta)**((1.0-self.beta)/self.beta) # profit coefficient
        self.ext = self.zeta*((1.0-self.beta)/self.beta) + 1.0 # externality factor
        self.mark_con = (1.0-self.beta)**(-1.0/self.beta) # Ystar/prof (control production)
        self.mark_mon = (2.0-self.beta)/(1.0-self.beta) # Yhat/prof (monopoly production)

        # regrid
        if grid:
            self.init_bins()

    # partially update params with keywords
    def update_params(self, grid=True, **kwargs):
        par_new = self.params.copy()
        for k, v in kwargs.items():
            if v is not None:
                par_new[k] = v
        self.set_params(params=par_new, grid=grid)

    # gridding and probability matrices
    def init_bins(self):
        # theta grid
        self.theta_mu = np.exp(self.ltheta_mu)
        self.ltheta_std = self.eps_std/np.sqrt(1-self.p**2)
        # self.ltheta1_range = np.sqrt(3)*ltheta_std # equation theta1 variance with long run theta variance
        self.ltheta_range = np.maximum(self.ltheta1_range, self.theta_range0*self.ltheta_std)
        self.ltheta_min = self.ltheta_mu - self.ltheta_range
        self.ltheta_max = self.ltheta_mu + self.ltheta_range
        self.ltheta_grid = np.linspace(self.ltheta_min, self.ltheta_max, self.N+1)
        self.ltheta_vec = 0.5*(self.ltheta_grid[:-1]+self.ltheta_grid[1:])
        self.theta_grid = np.exp(self.ltheta_grid)
        self.theta_vec = np.exp(self.ltheta_vec)

        # theta1 grid
        # theta1_range = np.sqrt(3.0)*theta_std # equating theta1_std and theta_std (long-run)
        self.ltheta1_min = self.ltheta_mu - self.ltheta1_range
        self.ltheta1_max = self.ltheta_mu + self.ltheta1_range
        self.ltheta1_grid = np.linspace(self.ltheta1_min, self.ltheta1_max, self.N1+1)
        self.ltheta1_vec = 0.5*(self.ltheta1_grid[:-1]+self.ltheta1_grid[1:])
        self.theta1_grid = np.exp(self.ltheta1_grid)
        self.theta1_vec = np.exp(self.ltheta1_vec)

        # theta1 distribution (log uniform)
        theta1_cmf0 = (self.ltheta1_grid-self.ltheta1_min)/(self.ltheta1_max-self.ltheta1_min)
        self.theta1_pmf = np.diff(theta1_cmf0)
        self.theta1_cmf = theta1_cmf0[1:]

        # theta[t,t+1] transition matrix
        self.ltheta_mid = self.ltheta_vec[:,None]
        self.ctrans_mat = stats.norm.cdf((self.ltheta_grid-(self.p*self.ltheta_mid+(1.0-self.p)*self.ltheta_mu))/self.eps_std)
        self.ctrans_mat[:,0] = 0.0
        self.ctrans_mat[:,-1] = 1.0
        self.ptrans_mat = np.diff(self.ctrans_mat, axis=1)

        # hazard rate
        # self.theta1_H1 = self.ltheta1_max - self.ltheta1_vec # only for uniform dist
        self.theta1_diff = np.diff(self.theta1_grid)
        self.theta1_pdf = self.theta1_pmf/self.theta1_diff
        self.theta1_H1 = (1-self.theta1_cmf)/(self.theta1_vec*self.theta1_pdf)

        # q grid
        self.q_grid = np.linspace(self.q_min, self.q_max, self.Q)
        self.q_grid = np.r_[self.q_min-self.q_grid[0], self.q_grid] # ensure that zero is on the vec
        self.q_vec = 0.5*(self.q_grid[:-1]+self.q_grid[1:])

        # time vector
        self.t_vec = np.arange(1, self.T+1)

        # mapping from theta1 to theta grid
        self.theta1_map = nd.project(self.theta_grid, xlo=self.theta1_grid[:-1], xhi=self.theta1_grid[1:])
        self.theta11_dst0 = self.theta1_map*self.theta1_pmf[:,None]
        self.theta12_dst0 = self.theta11_dst0[:,None,:]*np.eye(self.N)[None,:,:]
        self.theta1_dst0 = np.sum(self.theta1_map*self.theta1_pmf[:,None], axis=0)
        self.theta2_dst0 = self.theta1_dst0[None,:]*np.eye(self.N)

        # theta1 dist matrix
        self.th11_dist = np.zeros((self.T, self.N1, self.N))
        self.th11_dist[0,:,:] = self.theta11_dst0
        for t in range(1, self.T):
            self.th11_dist[t,:,:] = np.sum(self.th11_dist[t-1,:,:,None]*self.ptrans_mat[None,:,:], axis=1)

        self.th12_dist = np.zeros((self.T, self.N1, self.N, self.N))
        self.th12_dist[0,:,:,:] = self.theta12_dst0
        self.th12_dist[1:,:,:,:] = self.th11_dist[:-1,:,:,None]*self.ptrans_mat[None,None,:,:]

        # new theta dist matrix
        self.th1_dist = np.zeros((self.T, self.N))
        self.th1_dist[0,:] = self.theta1_dst0
        for t in range(1, self.T):
            self.th1_dist[t,:] = np.sum(self.th1_dist[t-1,:,None]*self.ptrans_mat, axis=0)

        self.th2_dist = np.zeros((self.T, self.N, self.N))
        self.th2_dist[0,:,:] = self.theta2_dst0
        self.th2_dist[1:,:,:] = self.th1_dist[:-1,:,None]*self.ptrans_mat[None,:,:]

        # initial q distribution
        blo = int(np.digitize(self.qbar0, self.q_grid, right=True))
        bhi = blo + 1
        qhi, qlo = self.q_vec[bhi], self.q_vec[blo]
        qx = (self.qbar0-qlo)/(qhi-qlo)
        self.qdist0 = np.zeros(self.Q)
        self.qdist0[bhi], self.qdist0[blo] = qx, 1 - qx

        # initial joint distribution
        self.dist0 = self.qdist0[:,None,None]*self.theta2_dst0[None,:,:]

        # otrack equilibrium store
        self.otrack = bn.Bundle()
        self.otrack.init = False
        self.otrack.qbar = np.zeros(self.T)
        self.otrack.l = np.zeros((self.T, self.N1, self.N, self.N))
        self.otrack.r = np.zeros((self.T, self.N1, self.N))
        self.otrack.Z = np.zeros(self.T)

        # ltrack equilibrium store
        self.ltrack = bn.Bundle()
        self.ltrack.init = False
        self.ltrack.qbar = np.zeros(self.T)
        self.ltrack.l = np.zeros((self.T, self.N, self.N))
        self.ltrack.r = np.zeros((self.T, self.N))
        self.ltrack.Z = np.zeros(self.T)

        # qtrack equilibrium store
        self.qtrack = bn.Bundle()
        self.qtrack.init = False
        self.qtrack.l = np.zeros((self.T, self.Q, self.N, self.N))
        self.qtrack.r = np.zeros((self.T, self.Q, self.N))
        self.qtrack.Z = np.zeros((self.T, self.Q, self.N, self.N))
        self.qtrack.dist = np.zeros((self.T, self.Q, self.N, self.N))

    # update eqvars with algo params
    def update(self, old, new):
        lo = np.maximum(self.val_min, (1-self.pct_max)*old)
        hi = np.minimum(self.val_max, (1+self.pct_max)*old)
        updated = self.dampen*old + (1-self.dampen)*new
        clamped = np.clip(updated, lo, hi)
        return clamped

    # find present value given params
    def present_value(self, x):
        disc = (1.0/self.R)**(np.arange(len(x)))
        return np.sum(x*disc)

    # w productivity function - CES between research stock and firm type
    def w_func(self, r, theta):
        if self.rho == 1.0:
            return r**self.alpha*theta**(1.0-self.alpha)
        else:
            return (self.alpha*r**(1.0-self.rho)+(1.0-self.alpha)*theta**(1.0-self.rho))**(1.0/(1.0-self.rho))

    ##
    ## solvers
    ##

    # initial guess
    def reset_otrack(self):
        self.otrack.init = True
        self.otrack.qbar[:] = 0.1
        self.otrack.l[:,:,:,:] = 0.1
        self.otrack.r[:,:,:] = 0.1
        self.otrack.Z[:] = 0.1

    def optimal(self, control=False, first=False, psub_wedge=False, reps=reps0, tol=tol0, bail=False, output=True, per=50, reset=False, save=True):
        # can we control quantities?
        pmark = self.mark_con if control else 1
        ymark = self.mark_con if control else self.mark_mon

        # informational rents
        theta1_H1 = np.zeros(self.N1) if first else self.theta1_H1

        # initial guess
        if reset or not self.otrack.init:
            self.reset_otrack()
        qbar = self.otrack.qbar.copy()
        l = self.otrack.l.copy()
        r = self.otrack.r.copy()
        Z = self.otrack.Z.copy()

        # theta[t,t+1] transitions
        ptrans_expect = lambda x: np.sum(x*self.ptrans_mat[None,None,:,:], axis=3)
        theta11_expect = lambda x: np.sum(x*self.th11_dist, axis=(1, 2))
        theta12_expect = lambda x: np.sum(x*self.th12_dist, axis=(1, 2, 3))
        shift_forward = lambda x: np.concat([x[1:,:,:], np.zeros((1, self.N1, self.N))], axis=0)

        # variable definitions
        # l: (t,th1,th-1,th)
        # r: (t,th1,th)
        # Z: (t,th1,th)

        # iterate
        conv = False
        for i in range(reps):
            # lagged R&D
            rp = np.concat([self.r0*np.ones((1, self.N1, self.N)), r[:-1,:]], axis=0)
            w = self.w_func(rp[:,:,:,None], self.theta_vec[None,None,None,:])
            lam = l*w

            # information quotients
            G = theta1_H1[None,:,None,None]*(self.p**(self.t_vec[:,None,None,None]-1))*(1.0+self.gamma)*(1.0-self.alpha)*(self.theta_vec[None,None,None,:]/w)**(1.0-self.rho)
            X = ((1-self.rho)/(1+self.gamma))*(G/(1+G))

            # update l
            l_ret = w*Z[:,None,None,None]/(1+G)
            l_new = (l_ret/self.kappal)**(1/self.gamma)

            # update r
            r_ret = (self.alpha/self.R)*Z[:,None,None]*ptrans_expect(((w/rp[:,:,:,None])**self.rho)*l*(1+X))
            r_fwd = (r_ret/(self.kappar))**(1/self.eta)
            r_new = shift_forward(r_fwd)

            # update Z
            ybar = ymark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta)
            Zn = np.concat([Z[1:], [0]], axis=0)
            Z_new = self.ext*ybar + ((1-self.delta)/self.R)*Zn

            # update qbar
            qbarp = np.concat([[self.qbar0], qbar[:-1]], axis=0)
            qbar_new = (1-self.delta)*qbarp + self.qboost*theta12_expect(lam)

            # differences
            qbar_err = np.max(np.abs(qbar_new-qbar))
            l_err = np.max(np.abs(l_new-l))
            r_err = np.max(np.abs(r_new-r))
            Z_err = np.max(np.abs(Z_new-Z))
            err = np.max([qbar_err, l_err, r_err, Z_err])

            # convergence or error
            if np.isnan(err):
                print(f'nan failure {i}')
                break
            elif bail and err > 100:
                print(f'bailing out {i}: {err}')
                break
            elif err < tol:
                if output:
                    print(f'converged {i}: {err}')
                conv = True
                break

            # progress
            if output and i % per == 0:
                qbarbar = np.mean(qbar)
                print(f'{i:4d}: qbar={qbarbar:9.4g} q={qbar_err:9.4g} l={l_err:9.4g} r={r_err:9.4g} Z={Z_err:9.4g} err={err:9.4g}')

            # update values
            qbar[:] = self.update(qbar, qbar_new)
            l[:,:,:] = self.update(l, l_new)
            r[:,:] = self.update(r, r_new)
            Z[:] = self.update(Z, Z_new)

        # save equilibrium variables
        if save and conv:
            self.otrack.qbar[:] = qbar
            self.otrack.l[:,:,:] = l
            self.otrack.r[:,:] = r
            self.otrack.Z[:] = Z

        # profit coefficient
        pibar = pmark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta)

        # output aggregates
        outp = ymark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta+1)
        piavg = pmark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta+1)

        # total cost
        phi = self.kappal*l**(1+self.gamma)/(1+self.gamma)
        rc = self.kappar*r**(1+self.eta)/(1+self.eta)

        # incentives relative to private incentives without subsidy
        Z0 = Z/self.ext # kill off externality
        Z0 /= ymark # add in monopoly term
        if psub_wedge:
            Z0 *= self.mark_con # (optionally) kill off monopoly term

        # compute wedges at optimum
        tax = 1.0 - self.kappal*l**self.gamma/(Z0[:,None,None,None]*w)
        mrc = self.kappar*rp**self.eta
        sub_base = (1/self.R)*Z0[:,None,None,None]*self.alpha*((w/rp[:,:,:,None])**self.rho)*l
        subt_fwd = 1.0 - ptrans_expect(sub_base*(1.0-tax))/mrc
        sub_fwd = 1.0 - ptrans_expect(sub_base)/mrc
        subt = shift_forward(subt_fwd)
        sub = shift_forward(sub_fwd)

        # marginal tax rates
        ptrans_expect1 = lambda x: np.sum(x*self.ptrans_mat[None,:,:], axis=2) # (th1,th-1,th)
        deprec = (1-self.delta)/self.R
        mt_exp = np.zeros((self.N1, self.N))
        mt = np.zeros((self.T, self.N1, self.N, self.N))
        mt[-1,:,:,:] = tax[-1,:,:,:]
        for t in reversed(range(self.T-1)):
            mt_exp = ptrans_expect1(pibar[t+1]*mt[t+1,:,:,:]+deprec*mt_exp[:,:,None])
            mt[t,:,:,:] = (tax[t,:,:,:]*Z0[t]-deprec*mt_exp[:,:,None])/pibar[t]

        # marginal subsidy rates
        dlam_dr = (1+self.gamma)/self.gamma*self.alpha*(w/rp[:,:,:,None])**self.rho*l
        mr_abs = sub*mrc + ptrans_expect(dlam_dr*tax*Z0[:,None,None,None])
        ms_fwd = mr_abs/mrc
        ms = shift_forward(ms_fwd)

        # compute welfare directly - informational term in eq (21)
        hazard = theta1_H1[None,:,None,None]
        pdisc = (self.p**(self.t_vec-1))[:,None,None,None]
        theta_all = self.theta_vec[None,None,None,:]
        util = hazard*pdisc*(1-self.alpha)*((theta_all/w)**(1-self.rho))*self.kappal*(l**(1+self.gamma))
        trans = util + phi

        # back out t=1 present value
        ptrans_expect1 = lambda x: np.sum(x*self.ptrans_mat[None,:,:], axis=2) # (th-1,th)
        V = np.zeros((self.T, self.N1, self.N, self.N))
        Vn = np.zeros((self.N1, self.N, self.N))
        for t in reversed(range(self.T)):
            V[t,:,:,:] = util[t,:,:,:] + (1/self.R)*ptrans_expect1(Vn)[:,None,:]
            Vn = V[t,:,:,:]
        V0 = np.sum(self.theta12_dst0*V[0,:,:,:], axis=(1, 2))/np.sum(self.theta12_dst0, axis=(1, 2))

        # compute marginals from transfers
        dtrans_dtheta = nd.deriv(trans, self.theta_vec, axis=3)
        dlam_dtheta = nd.deriv(lam, self.theta_vec, axis=3)
        # mt_new = 1 - (1-1/self.R)*dtrans_dtheta/(prof[:,None,None,None]*dlam_dtheta)
        mt_norm = 1 - dtrans_dtheta/(Z0[:,None,None,None]*dlam_dtheta)
        mt_new = dtrans_dtheta/(pibar[:,None,None,None]*dlam_dtheta)

        # find current q value from lambda choice
        qc = (1-self.delta)*self.q_vec[None,:,None,None,None] + lam[:,None,:,:,:]
        qcproj = np.moveaxis(nd.project(self.q_grid, xp=qc, axis=1), -1, 2)

        # calculate qdist through time
        dist = np.zeros((self.T, self.Q, self.N1, self.N, self.N)) # this is a whopper
        dist[0,:,:,:,:] = self.qdist0[:,None,None,None]*self.theta12_dst0[None,:,:,:]
        for t in range(1, self.T):
            distq = np.sum(dist[t-1,:,None,:,:,:]*qcproj[t-1,:,:,:,:,:], axis=0)
            distt = np.sum(distq, axis=2)
            dist[t,:,:,:,:] = distt[:,:,:,None]*self.ptrans_mat[None,None,:,:]

        # some q based numbers
        pi = pibar[:,None,None,None,None]*qc

        # time series stats
        tstats = pd.DataFrame(index=self.t_vec)
        tstats['t'] = self.t_vec
        tstats['qbar'] = qbar
        tstats['outp'] = outp
        tstats['pi'] = piavg

        # averages - (t,th1,th-1,th)
        tstats['lam'] = theta12_expect(lam)
        tstats['l'] = theta12_expect(l)
        tstats['r'] = theta11_expect(r)
        tstats['phi'] = theta12_expect(phi)
        tstats['rc'] = theta11_expect(rc)
        tstats['tax'] = theta12_expect(tax)
        tstats['sub'] = theta11_expect(sub)
        tstats['subt'] = theta11_expect(subt)
        tstats['mt'] = theta12_expect(mt)
        tstats['ms'] = theta11_expect(ms)
        tstats['util'] = theta12_expect(util)
        tstats['trans'] = theta12_expect(trans)

        # welfare
        tstats['cons'] = tstats['outp'] - tstats['rc'] - tstats['trans']
        welfare = self.present_value(tstats['cons'])

        return bn.local_bundle([
            'qbar', 'l', 'r', 'w', 'lam', 'Z', 'rc', 'phi', 'dist', 'tax', 'sub', 'subt', 'mt', 'ms', 'mt_new', 'mt_norm', 'trans', 'conv', 'welfare', 'V', 'V0', 'pi'
        ], locals()), tstats

    # initial guess
    def reset_ltrack(self):
        self.ltrack.init = True
        self.ltrack.qbar[:] = 0.1
        self.ltrack.l[:,:,:] = 0.1
        self.ltrack.r[:,:] = 0.1
        self.ltrack.Z[:] = 0.1

    def get_ltrack(self):
        return self.ltrack.deepcopy()

    def set_ltrack(self, lt):
        self.ltrack = lt

    def equilibrium(self, tau=tau_base, sub=sub_base, control=False, recoup='time', reps=reps0, tol=tol0, bail=False, output=True, per=50, reset=False, save=True):
        # can we control quantities?
        pmark = self.mark_con if control else 1
        ymark = self.mark_con if control else self.mark_mon

        # time variation
        tau = tau*np.ones_like(self.t_vec)
        sub = sub*np.ones_like(self.t_vec)

        # initial guess
        if reset or not self.ltrack.init:
            self.reset_ltrack()
        qbar = self.ltrack.qbar.copy()
        l = self.ltrack.l.copy()
        r = self.ltrack.r.copy()
        Z = self.ltrack.Z.copy()

        # theta[t,t+1] transitions
        ptrans_expect = lambda x: np.sum(x*self.ptrans_mat[None,:,:], axis=2)
        theta1_expect = lambda x: np.sum(x*self.th1_dist, axis=1)
        theta2_expect = lambda x: np.sum(x*self.th2_dist, axis=(1, 2))
        shift_forward = lambda x: np.concat([x[1:,:], np.zeros((1, self.N))], axis=0)

        # variable definitions
        # l: (t,th-1,th)
        # r: (t,th)
        # Z: (t,th)

        # iterate
        conv = False
        for i in range(reps):
            # lagged R&D
            rp = np.concat([self.r0*np.ones((1, self.N)), r[:-1,:]], axis=0)
            w = self.w_func(rp[:,:,None], self.theta_vec[None,None,:])
            lam = l*w

            # update l
            l_ret = w*Z[:,None,None]
            l_new = (l_ret/self.kappal)**(1/self.gamma)

            # update r
            subp = np.concat([np.zeros(1), sub[:-1]])
            r_ret = (self.alpha/self.R)*Z[:,None]*ptrans_expect(((w/rp[:,:,None])**self.rho)*l)
            r_fwd = (r_ret/((1-subp[:,None])*self.kappar))**(1/self.eta)
            r_new = shift_forward(r_fwd)

            # update Z
            pibar = pmark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta)
            Zn = np.concat([Z[1:], [0]], axis=0)
            Z_new = (1-tau)*pibar + ((1-self.delta)/self.R)*Zn

            # update qbar
            qbarp = np.concat([[self.qbar0], qbar[:-1]], axis=0)
            qbar_new = (1-self.delta)*qbarp + self.qboost*theta2_expect(lam)

            # differences
            qbar_err = np.max(np.abs(qbar_new-qbar))
            l_err = np.max(np.abs(l_new-l))
            r_err = np.max(np.abs(r_new-r))
            Z_err = np.max(np.abs(Z_new-Z))
            err = np.max([qbar_err, l_err, r_err, Z_err])

            # convergence or error
            if np.isnan(err):
                print(f'nan failure {i}')
                break
            elif bail and err > 100:
                print(f'bailing out {i}: {err}')
                break
            elif err < tol:
                if output:
                    print(f'converged {i}: {err}')
                conv = True
                break

            # progress
            if output and i % per == 0:
                qbarbar = np.mean(qbar[:sim_cut])
                print(f'{i:4d}: qbar={qbarbar:9.4g} q={qbar_err:9.4g} l={l_err:9.4g} r={r_err:9.4g} Z={Z_err:9.4g}')

            # update values
            qbar[:] = self.update(qbar, qbar_new)
            l[:,:,:] = self.update(l, l_new)
            r[:,:] = self.update(r, r_new)
            Z[:] = self.update(Z, Z_new)

        # save equilibrium variables
        if save and conv:
            self.ltrack.qbar[:] = qbar
            self.ltrack.l[:,:,:] = l
            self.ltrack.r[:,:] = r
            self.ltrack.Z[:] = Z

        # output aggregates
        outp = ymark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta+1)
        piavg = pmark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta+1)

        # total cost
        phi = self.kappal*l**(1+self.gamma)/(1+self.gamma)
        rc = self.kappar*r**(1+self.eta)/(1+self.eta)

        # tax totals
        ptax = tau*pibar
        rsub = sub[:,None]*rc

        # find current q value from lambda choice
        qc = (1-self.delta)*self.q_vec[None,:,None,None] + lam[:,None,:,:]
        qcproj = np.moveaxis(nd.project(self.q_grid, xp=qc, axis=1), -1, 2)

        # calculate qdist through time
        dist = np.zeros((self.T, self.Q, self.N, self.N))
        dist[0,:,:,:] = self.qdist0[:,None,None]*self.theta2_dst0[None,:,:]
        for t in range(1, self.T):
            distq = np.sum(dist[t-1,:,None,:,:]*qcproj[t-1,:,:,:,:], axis=0)
            distt = np.sum(distq, axis=1)
            dist[t,:,:,:] = distt[:,:,None]*self.ptrans_mat[None,:,:]

        # back out t=1 present value
        piq = pibar[:,None,None,None]*qc
        ptaxq = tau[:,None,None,None]*piq
        u = (piq-ptaxq) - phi[:,None,:,:] - (rc-rsub)[:,None,None,:]
        ptrans_expect1 = lambda x: np.sum(x*self.ptrans_mat[None,:,:], axis=2) # (th-1,th)
        V = np.zeros((self.T, self.Q, self.N, self.N))
        Vn = np.zeros((self.Q, self.N, self.N))
        for t in reversed(range(self.T)):
            Vi = nd.interp(qc[t,:,:,:], self.q_vec, Vn, axis=0) # project qc onto qgrid
            Vt = ptrans_expect1(Vi) # expectation over theta(t+1)
            V[t,:,:,:] = u[t,:,:,:] + (1/self.R)*Vt[:,None,:]
            Vn = V[t,:,:,:]
        V0 = np.sum(self.dist0*V[0,:,:,:], axis=(0, 1))/np.sum(self.dist0, axis=(0, 1))

        # time series stats
        tstats = pd.DataFrame(index=self.t_vec)
        tstats['t'] = self.t_vec
        tstats['qbar'] = qbar
        tstats['outp'] = outp
        tstats['pi'] = piavg

        # averages - (t,th-1,th)
        tstats['lam'] = theta2_expect(lam)
        tstats['l'] = theta2_expect(l)
        tstats['r'] = theta1_expect(r)
        tstats['phi'] = theta2_expect(phi)
        tstats['rc'] = theta1_expect(rc)
        tstats['ptax'] = ptax
        tstats['rsub'] = theta1_expect(rsub)
        tstats['util'] = (tstats['pi']-tstats['ptax']) - (tstats['rc']-tstats['rsub']) - tstats['phi']

        # recoup lump sum
        if recoup == 'full':
            Vmin = np.min(V0)
            tstats['lump'] = np.r_[Vmin, np.zeros(self.T-1)]
        elif recoup == 'time':
            Vmin = np.min(V, axis=(1, 2, 3))
            tstats['lump'] = Vmin - (1/self.R)*np.r_[Vmin[1:], 0.0]
        else:
            tstats['lump'] = 0

        # transfers
        tstats['trans'] = (tstats['pi'] - tstats['ptax']) - (tstats['rc'] - tstats['rsub']) - tstats['lump']

        # welfare
        tstats['cons'] = tstats['outp'] - tstats['rc'] - tstats['trans']
        welfare = self.present_value(tstats['cons'])

        return bn.local_bundle([
            'qbar', 'l', 'r', 'w', 'lam', 'Z', 'rc', 'phi', 'piq', 'welfare', 'i', 'conv', 'tol', 'err', 'i'
        ], locals()), tstats

    # initial guess
    def reset_qtrack(self):
        self.qtrack.init = True
        self.qtrack.l[:,:,:,:] = 0.1
        self.qtrack.r[:,:,:] = 0.1
        self.qtrack.Z[:,:,:,:] = 0.1
        self.qtrack.dist[:,:,:,:] = self.dist0[None,:,:,:]

    def get_qtrack(self):
        return self.qtrack.deepcopy()

    def set_qtrack(self, qt):
        self.qtrack = qt

    # functional formulation - dx is marginal tax rate, tx is total tax payments
    # recoup in ['none', 'time', 'full']
    def equilibrium_func(self, dtau=dtau_base, ttau=ttau_base, dsub=dsub_base, tsub=tsub_base, dcross=dcross_base, control=False, recoup='time', reps=reps0, tol=tol0, bail=False, output=True, per=50, reset=False, save=True, welfare=None):
        # can we control quantities?
        pmark = self.mark_con if control else 1
        ymark = self.mark_con if control else self.mark_mon

        # initial guess
        if reset or not self.qtrack.init:
            self.reset_qtrack()
        l = self.qtrack.l.copy()
        r = self.qtrack.r.copy()
        Z = self.qtrack.Z.copy()
        dist = self.qtrack.dist.copy()

        # expectations
        ptrans_expect = lambda x: np.sum(x*self.ptrans_mat[None,None,:,:], axis=3) # (th-1,th)
        dist_expect = lambda x: np.sum(x*dist, axis=(1, 2, 3)) # (t,q-1,th-1,th)

        # variable definitions
        # l: (t,q-1,th-1,th)
        # r: (t,q,th)
        # Z: (t,q-1,th-1,th)
        # dist: (t,q-1,th-1,th)

        # iterate
        conv = False
        for i in range(reps):
            # lam choice
            rp = np.concat([self.r0*np.ones((1, self.Q, self.N)), r[:-1,:,:]], axis=0)
            w = self.w_func(rp[:,:,:,None], self.theta_vec[None,None,None,:])
            lam = l*w

            # update l1
            l_ret = w*Z
            l_new = (l_ret/self.kappal)**(1/self.gamma)

            # profit
            qc = (1-self.delta)*self.q_vec[None,:,None,None] + lam
            qbar = self.qboost*dist_expect(qc)
            pibar = pmark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta)
            pi = pibar[:,None,None,None]*qc

            # subsidy rates
            rc = self.kappar*(r**(1+self.eta))/(1+self.eta)
            rcp = self.kappar*(rp**(1+self.eta))/(1+self.eta)
            drcp = self.kappar*(rp**(self.eta))
            m_sub = dsub(rc)
            m_subp = np.concat([np.zeros((1, self.Q, self.N)), m_sub[:-1,:,:]], axis=0)
            m_cross = dcross(pi, rcp[:,:,:,None])

            # update r
            r_ret = (self.alpha/self.R)*ptrans_expect(((w/rp[:,:,:,None])**self.rho)*l*Z-m_cross*drcp[:,:,:,None])
            r_fwd = (np.maximum(0, r_ret)/(self.kappar*(1-m_sub)))**(1/self.eta)
            r_new = np.concat([r_fwd[1:,:,:], np.zeros((1, self.Q, self.N))], axis=0)

            # after tax
            m_tau = dtau(pi, rcp[:,:,:,None])
            tpibar = (1-m_tau)*pibar[:,None,None,None]

            # update Z
            Zn = np.concat([Z[1:,:,:,:], np.zeros((1, self.Q, self.N, self.N))], axis=0)
            Zi = nd.interp(qc, self.q_vec, Zn, axis=1)
            Z_new = tpibar + ((1-self.delta)/self.R)*ptrans_expect(Zi)[:,:,None,:]

            # update dist
            qcproj = np.moveaxis(nd.project(self.q_grid, xp=qc, axis=1), -1, 2)
            dist1 = np.sum(dist[:,:,None,:,:]*qcproj, axis=1) # project qc on qgrid
            dist2 = np.sum(dist1, axis=2) # average out theta(t-1)
            dist3 = dist2[:,:,:,None]*self.ptrans_mat[None,None,:,:] # theta(t-1) to theta(t) probability
            dist_new = np.concat([self.dist0.reshape((1, self.Q, self.N, self.N)), dist3[:-1,:,:,:]], axis=0)

            # differences
            l_err = np.max(np.abs(l_new-l))
            r_err = np.max(np.abs(r_new-r))
            Z_err = np.max(np.abs(Z_new-Z))
            dist_err = np.max(np.abs(dist_new-dist))
            err = np.max([l_err, r_err, Z_err, dist_err])

            # convergence
            if np.isnan(err):
                if output:
                    print(f'nan failure {i}')
                break
            elif bail and err > 1000:
                if output:
                    print(f'out of bounds {i}: {err}')
                break
            elif err < tol:
                if output:
                    print(f'converged {i}: {err}')
                conv = True
                break

            # progress
            if output and i % per == 0:
                qbarbar = np.mean(qbar[:sim_cut])
                print(f'{i:4d}: qbar={qbarbar:9.4g} d={dist_err:9.4g} l={l_err:9.4g} r={r_err:9.4g} Z={Z_err:9.4g} err={err:9.4g}')

            # update values
            l[:,:,:,:] = self.update(l, l_new)
            r[:,:,:] = self.update(r, r_new)
            Z[:,:,:] = self.update(Z, Z_new)
            dist[:,:,:] = self.dampen*dist + (1-self.dampen)*dist_new

        # save equilibrium variables
        if save and conv:
            self.qtrack.l[:,:,:,:] = l
            self.qtrack.r[:,:,:] = r
            self.qtrack.Z[:,:,:,:] = Z
            self.qtrack.dist[:,:,:,:] = dist

        # output aggregates
        outp = ymark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta+1)
        piavg = pmark*self.psi*qbar**(self.zeta*(1-self.beta)/self.beta+1)

        # allocation
        rc = self.kappar*(r**(1+self.eta))/(1+self.eta)
        phi = self.kappal*(l**(1+self.gamma))/(1+self.gamma)

        # tax totals
        ptax = ttau(pi, rcp[:,:,:,None])
        rsub = tsub(rc)

        # back out t=1 present value
        u = (pi-ptax) - phi - (rc-rsub)[:,:,None,:]
        ptrans_expect1 = lambda x: np.sum(x*self.ptrans_mat[None,:,:], axis=2) # (th-1,th)
        V = np.zeros((self.T, self.Q, self.N, self.N))
        Vn = np.zeros((self.Q, self.N, self.N))
        for t in reversed(range(self.T)):
            Vi = nd.interp(qc[t,:,:,:], self.q_vec, Vn, axis=0)
            V[t,:,:,:] = u[t,:,:,:] + (1/self.R)*ptrans_expect1(Vi)[:,None,:]
            Vn = V[t,:,:,:]
        V0 = np.sum(self.dist0*V[0,:,:,:], axis=(0, 1))/np.sum(self.dist0, axis=(0, 1))

        # time series stats
        tstats = pd.DataFrame(index=self.t_vec)
        tstats['t'] = self.t_vec
        tstats['qbar'] = qbar
        tstats['outp'] = outp
        tstats['prof'] = piavg

        # averages - (t,q-1,th-1,th)
        tstats['lam'] = dist_expect(lam)
        tstats['l'] = dist_expect(l)
        tstats['phi'] = dist_expect(phi)
        tstats['r'] = dist_expect(r[:,:,None,:])
        tstats['rc'] = dist_expect(rc[:,:,None,:])
        tstats['pi'] = dist_expect(pi)
        tstats['ptax'] = dist_expect(ptax)
        tstats['rsub'] = dist_expect(rsub[:,:,None,:])
        tstats['util'] = dist_expect(u)

        # recoup lump sum
        if recoup == 'full':
            Vmin = np.min(V0)
            tstats['lump'] = np.r_[Vmin, np.zeros(self.T-1)]
        elif recoup == 'time':
            Vmin = np.min(V, axis=(1, 2, 3))
            tstats['lump'] = Vmin - (1/self.R)*np.r_[Vmin[1:], 0.0]
        else:
            tstats['lump'] = 0

        # transfers
        tstats['trans'] = (tstats['prof'] - tstats['ptax']) - (tstats['rc'] - tstats['rsub']) - tstats['lump']

        # welfare
        tstats['cons'] = tstats['outp'] - tstats['rc'] - tstats['trans']
        welfare = self.present_value(tstats['cons'])

        # construct equilibrium
        eq = bn.local_bundle([
            'qbar', 'l', 'r', 'dist', 'w', 'lam', 'pi', 'qc', 'Z', 'Zi', 'rc', 'phi', 'welfare', 'i', 'conv', 'tol', 'l_err', 'r_err', 'Z_err', 'dist_err', 'err', 'm_tau', 'm_sub', 'u', 'V', 'V0', 'qc', 'qcproj', 'ptax', 'rsub', 'rcp'
        ], locals())

        return eq, tstats

    def equilibrium_lin1(self, tau, sub, **kwargs):
        dtau, ttau, dsub, tsub = policy_lin(tau, sub)
        return self.equilibrium_func(dtau, ttau, dsub, tsub, **kwargs)

    def equilibrium_lin(self, tau, sub, **kwargs):
        return self.equilibrium(tau, sub, **kwargs)

    def equilibrium_lindec(self, tau, tdec, sub, sdec, **kwargs):
        tau = tau*np.exp(-tdec*self.t_vec)
        sub = sub*np.exp(-sdec*self.t_vec)
        return self.equilibrium(tau, sub, **kwargs)

    def equilibrium_lindecint(self, tau, tdec0, tdec1, sub, sdec0, sdec1, **kwargs):
        tau = tdec0 + tau*np.exp(-tdec1*self.t_vec)
        sub = sdec0 + sub*np.exp(-sdec1*self.t_vec)
        return self.equilibrium(tau, sub, **kwargs)

    def equilibrium_linx(self, tau0, tau1, sub0, **kwargs):
        dtau, ttau, dsub, tsub, dcross = policy_linx(tau0, tau1, sub0)
        return self.equilibrium_func(dtau, ttau, dsub, tsub, dcross, **kwargs)

    def equilibrium_linxdec(self, tau0, tau1, tdec, sub0, sdec, **kwargs):
        tau0 = tau0*np.exp(-tdec*self.t_vec)[:,None,None,None]
        tau1 = tau1*np.exp(-tdec*self.t_vec)[:,None,None,None]
        sub0 = sub0*np.exp(-sdec*self.t_vec)[:,None,None]
        return self.equilibrium_linx(tau0, tau1, sub0, **kwargs)

    def equilibrium_hsv(self, tau0, tau1, sub0, sub1, tau2, sub2, **kwargs):
        dtau, ttau, dsub, tsub = policy_hsv(tau0, tau1, sub0, sub1, tau2, sub2)
        return self.equilibrium_func(dtau, ttau, dsub, tsub, **kwargs)

    def equilibrium_hsvx(self, tau0, tau1, tau2, tau3, sub0, sub1, sub2, **kwargs):
        dtau, ttau, dsub, tsub, dcross = policy_hsvx(tau0, tau1, tau2, tau3, sub0, sub1, sub2)
        return self.equilibrium_func(dtau, ttau, dsub, tsub, dcross, **kwargs)

    def equilibrium_linhsv(self, tau, sub0, sub1, sub2, **kwargs):
        dtau, ttau, dsub, tsub = policy_linhsv(tau, sub0, sub1, sub2)
        return self.equilibrium_func(dtau, ttau, dsub, tsub, **kwargs)

    def equilibrium_hsvlin(self, tau0, tau1, tau2, sub, **kwargs):
        dtau, ttau, dsub, tsub = policy_hsvlin(tau0, tau1, tau2, sub)
        return self.equilibrium_func(dtau, ttau, dsub, tsub, **kwargs)

    ##
    ## simulate!
    ##

    def generate_panel(self, eq, NF=NF0, seed=SEED):
        state = np.random.RandomState(seed)

        # transition probabilities
        trans_cmf = self.ctrans_mat[:,1:]

        # initial state
        itheta = nd.random_vec(self.theta1_cmf, NF, state=state)
        qval = self.qbar0*np.ones(NF)

        # calculate and store
        qbar = eq['qbar'][0]
        theta = self.theta_vec[itheta]
        rnd = eq['rc'][0,itheta]
        lam = eq['lam'][0,itheta,itheta] # here the t-1 (=0) dimension should be constant
        qval = (1.0-self.delta)*qval + lam

        firms = []
        firms.append({
            'itheta': itheta,
            'qbar': qbar,
            'rnd': rnd,
            'lam': lam,
            'qval': qval
        })

        # roll through periods (should we include last period where r=0?)
        for t in range(1, self.T):
            # increment theta
            itheta_next = nd.random_mat(trans_cmf, itheta, state=state)

            # calculate and store
            qbar = eq['qbar'][t]
            theta_next = self.theta_vec[itheta_next]
            rnd = eq['rc'][t,itheta_next]
            lam = eq['lam'][t,itheta,itheta_next]
            qval = (1.0-self.delta)*qval + lam

            firms.append({
                'itheta': itheta_next,
                'theta': theta_next,
                'qbar': qbar,
                'rnd': rnd,
                'lam': lam,
                'qval': qval
            })

            # update theta
            itheta = itheta_next

        # stack it up
        panel = stack_dicts(firms, agg_func=np.vstack)

        # cut off end history
        panel = {k: v[:sim_cut,:] for k, v in panel.items()}

        return panel

    def calc_moments(self, panel, infer=False, nthread=5, output=False, save=False, frame=False, seed=SEED, target=target_default_fname):
        _, nF = panel['itheta'].shape

        # intermediate values
        panel['sales'] = (self.psi/self.beta)*panel['qbar']**(self.zeta*(1.0-self.beta)/self.beta)*panel['qval']
        panel['income'] = self.psi*panel['qbar']**(self.zeta*(1.0-self.beta)/self.beta)*panel['qval']
        panel['growth_dhs'] = 2.0*(panel['sales'][1:,:]-panel['sales'][:-1,:])/(panel['sales'][1:,:]+panel['sales'][:-1,:])
        panel['rndi'] = panel['rnd']/panel['sales']
        panel['lsales'] = np.log(panel['sales'])
        panel['rnd1'] = np.vstack([np.nan*np.ones(nF), panel['rnd'][:-1,:]])
        panel['lam1'] = np.vstack([np.nan*np.ones(nF), panel['lam'][:-1,:]])
        panel['lrnd'] = np.log(panel['rnd'])
        panel['lrnd1'] = np.log(panel['rnd1'])
        panel['llam'] = np.log(panel['lam'])
        panel['llam1'] = np.log(panel['lam1'])

        # regression
        data = pd.DataFrame({k: panel[k].flatten() for k in ['llam', 'lrnd1']})
        ret = smf.ols('llam ~ lrnd1', data=data.dropna()).fit()
        patrnd_elast = ret.params['lrnd1']

        # young and old
        pats_all = panel['lam']
        pats_1 = panel['lam'][0,:]
        pats_p = panel['lam'][1:,:]
        pats_yng = panel['lam'][:age_cut,:]
        pats_old = panel['lam'][age_cut:,:]

        # true moments
        pats_within_cov = np.mean(np.std(pats_all, axis=0)/np.mean(pats_all, axis=0))
        pats_across_all_cov = np.mean(np.std(pats_all, axis=1)/np.mean(pats_all, axis=1))
        pats_across_yng_cov = np.mean(np.std(pats_yng, axis=1)/np.mean(pats_yng, axis=1))
        pats_across_old_cov = np.mean(np.std(pats_old, axis=1)/np.mean(pats_old, axis=1))
        gsales_mean = np.mean(panel['growth_dhs'])
        rndi_mean = np.mean(panel['rndi'])
        rndi_median = np.median(panel['rndi'])
        ros_median = self.beta
        pats_ratio = np.mean(pats_yng)/np.mean(pats_old)

        # log stdev
        pats_within_lstd = np.mean(np.std(log(pats_all), axis=0))
        pats_across_all_lstd = np.mean(np.std(log(pats_all), axis=1))
        pats_across_yng_lstd = np.mean(np.std(log(pats_yng), axis=1))
        pats_across_old_lstd = np.mean(np.std(log(pats_old), axis=1))

        # non-targeted moments
        acut1 = 3
        acut2 = 5
        acut3 = 10

        # 2 - rndi by age
        rndi_age1_yng = np.mean(panel['rndi'][:acut1,:])
        rndi_age1_old = np.mean(panel['rndi'][acut1:,:])
        rndi_age2_yng = np.mean(panel['rndi'][:acut2,:])
        rndi_age2_old = np.mean(panel['rndi'][acut2:,:])
        rndi_age3_yng = np.mean(panel['rndi'][:acut3,:])
        rndi_age3_old = np.mean(panel['rndi'][acut3:,:])

        # 10 - sales by age - these need to be normalized by mean sales
        sales_mean = np.mean(panel['sales'])
        sales_age1_yng = np.mean(panel['sales'][:acut1,:])/sales_mean
        sales_age1_old = np.mean(panel['sales'][acut1:,:])/sales_mean
        sales_age2_yng = np.mean(panel['sales'][:acut2,:])/sales_mean
        sales_age2_old = np.mean(panel['sales'][acut2:,:])/sales_mean
        sales_age3_yng = np.mean(panel['sales'][:acut3,:])/sales_mean
        sales_age3_old = np.mean(panel['sales'][acut3:,:])/sales_mean

        # 11 - sales growth by age (mean)
        gsales_age1_yng_mean = np.mean(panel['growth_dhs'][:acut1,:])
        gsales_age1_old_mean = np.mean(panel['growth_dhs'][acut1:,:])
        gsales_age2_yng_mean = np.mean(panel['growth_dhs'][:acut2,:])
        gsales_age2_old_mean = np.mean(panel['growth_dhs'][acut2:,:])
        gsales_age3_yng_mean = np.mean(panel['growth_dhs'][:acut3,:])
        gsales_age3_old_mean = np.mean(panel['growth_dhs'][acut3:,:])

        # 11 - sales growth by age (median)
        gsales_age1_yng_median = np.median(panel['growth_dhs'][:acut1,:])
        gsales_age1_old_median = np.median(panel['growth_dhs'][acut1:,:])
        gsales_age2_yng_median = np.median(panel['growth_dhs'][:acut2,:])
        gsales_age2_old_median = np.median(panel['growth_dhs'][acut2:,:])
        gsales_age3_yng_median = np.median(panel['growth_dhs'][:acut3,:])
        gsales_age3_old_median = np.median(panel['growth_dhs'][acut3:,:])

        # 11 - persistence of firm growth
        gsales_lag_cov = np.cov([panel['growth_dhs'][:-1,:].flatten(),panel['growth_dhs'][1:,:].flatten()])[0,1]
        gsales_lag_cor = np.corrcoef([panel['growth_dhs'][:-1,:].flatten(),panel['growth_dhs'][1:,:].flatten()])[0,1]

        # rndi quantiles
        sales_growth_5yr = 2.0*(panel['sales'][5,:]-panel['sales'][0,:])/(panel['sales'][5,:]+panel['sales'][0,:])
        sales_growth_5yr_pct90 = np.quantile(sales_growth_5yr, 0.9)
        rndi_bottom90 = panel['rndi'][:,sales_growth_5yr<sales_growth_5yr_pct90].mean()
        rndi_top10 = panel['rndi'][:,sales_growth_5yr>=sales_growth_5yr_pct90].mean()
        sales_growth_bottom90 = panel['growth_dhs'][:,sales_growth_5yr<sales_growth_5yr_pct90].mean()
        sales_growth_top10 = panel['growth_dhs'][:,sales_growth_5yr>=sales_growth_5yr_pct90].mean()

        # compile
        mmt_sim = OrderedDict([
            ('rnd_elasticity',          patrnd_elast),
            ('rndi_mean',               rndi_mean),
            ('rndi_median',             rndi_median),
            ('sales_growth',            gsales_mean),
            ('patqual_within_cov',      pats_within_cov),
            ('patqual_across_all_cov',  pats_across_all_cov),
            ('patqual_across_yng_cov',  pats_across_yng_cov),
            ('patqual_across_old_cov',  pats_across_old_cov),
            ('patqual_ratio',           pats_ratio),
            ('spillover_reg',           np.nan),
            ('subsidies_reg',           np.nan),
            ('rndi_age1_yng',           rndi_age1_yng),
            ('rndi_age1_old',           rndi_age1_old),
            ('rndi_age2_yng',           rndi_age2_yng),
            ('rndi_age2_old',           rndi_age2_old),
            ('rndi_age3_yng',           rndi_age3_yng),
            ('rndi_age3_old',           rndi_age3_old),
            ('sales_age1_yng',          sales_age1_yng),
            ('sales_age1_old',          sales_age1_old),
            ('sales_age2_yng',          sales_age2_yng),
            ('sales_age2_old',          sales_age2_old),
            ('sales_age3_yng',          sales_age3_yng),
            ('sales_age3_old',          sales_age3_old),
            ('gsales_age1_yng_mean',    gsales_age1_yng_mean),
            ('gsales_age1_old_mean',    gsales_age1_old_mean),
            ('gsales_age2_yng_mean',    gsales_age2_yng_mean),
            ('gsales_age2_old_mean',    gsales_age2_old_mean),
            ('gsales_age3_yng_mean',    gsales_age3_yng_mean),
            ('gsales_age3_old_mean',    gsales_age3_old_mean),
            ('gsales_age1_yng_median',  gsales_age1_yng_median),
            ('gsales_age1_old_median',  gsales_age1_old_median),
            ('gsales_age2_yng_median',  gsales_age2_yng_median),
            ('gsales_age2_old_median',  gsales_age2_old_median),
            ('gsales_age3_yng_median',  gsales_age3_yng_median),
            ('gsales_age3_old_median',  gsales_age3_old_median),
            ('gsales_lag_cov',          gsales_lag_cov),
            ('gsales_lag_cor',          gsales_lag_cor),
            ('pats_within_lstd',        pats_within_lstd),
            ('pats_across_all_lstd',    pats_across_all_lstd),
            ('pats_across_yng_lstd',    pats_across_yng_lstd),
            ('pats_across_old_lstd',    pats_across_old_lstd),
            ('rndi_bottom90',           rndi_bottom90),
            ('rndi_top10',              rndi_top10),
            ('sales_growth_bottom90',   sales_growth_bottom90),
            ('sales_growth_top10',      sales_growth_top10),
        ])

        # indirect inference
        if infer:
            mmt_sim['spillover_reg'] = self.calc_reg_zeta(nthread=nthread)
            mmt_sim['subsidies_reg'] = self.calc_reg_subs(nthread=nthread)

        if frame or output or save:
            mmt_dat = load_toml(target)

        if output or save:
            msg = '{:>5s}  {:25s} {:>10s} {:>10s}\n'.format('#','name','data','simul')
            for (i, k) in enumerate(mmt_names):
                msg += '{:5d}  {:25s} {:10f} {:10f}\n'.format(i+1, k, mmt_dat[k], mmt_sim[k])
            if output:
                print(msg)

        if save:
            f = open('output/moments.txt', 'w+')
            f.write(msg)

        if frame:
            return pd.DataFrame({'data': mmt_dat, 'simul': mmt_sim}, index=pd.Index(mmt_sim))
        else:
            return mmt_sim

    def calc_reg_zeta(self, nind=5, nthread=5, NF=1024, seed=SEED, output=False):
        # generate even range around kappa
        pct_idx = np.linspace(0.0, 1.0, nind+2)[1:-1]
        lqb_vec = stats.norm.ppf(pct_idx, loc=0.0, scale=0.2)
        qbs_vec = np.exp(lqb_vec)
        par_vec = [dict(self.params, qboost=qb) for qb in qbs_vec]

        # loop over qboost values
        inp_vec = [(par, NF, seed) for par in par_vec]
        with Pool(nthread) as p:
            fpanel = pd.concat(p.starmap(regression_op, inp_vec, chunksize=1), axis=0)

        # add in logs
        log_cols = ['qbar', 'sales', 'rnd', 'rnd_cum', 'rnd_lag', 'rnd_avg', 'rnd_cum_avg']
        lpanel = fpanel[log_cols].apply(log).add_prefix('log_')
        fpanel = pd.concat([fpanel, lpanel], axis=1).dropna()

        # regress
        ret_rev = smf.ols('log_sales ~ log_rnd + log_rnd_avg', data=fpanel).fit()

        if output:
            return fpanel, ret_rev
        else:
            return ret_rev.params['log_rnd_avg']

    def calc_reg_subs(self, nind=5, nthread=5, NF=1024, seed=SEED, output=False):
        # generate even range around kappa
        kappar0 = self.kappar
        pct_idx = np.linspace(0.0, 1.0, nind+2)[1:-1]
        krs_vec = stats.norm.ppf(pct_idx, loc=kappar0, scale=0.3*kappar0)
        par_vec = [dict(self.params, kappar=kr) for kr in krs_vec]

        # loop over kappa values
        inp_vec = [(par, NF, seed) for par in par_vec]
        with Pool(nthread) as p:
            fpanel = pd.concat(p.starmap(regression_op, inp_vec, chunksize=1), axis=0)

        # add in logs
        log_cols = ['kappar', 'sales', 'rnd', 'rnd_cum', 'rnd_lag', 'rnd_avg', 'rnd_cum_avg']
        lpanel = fpanel[log_cols].apply(log).add_prefix('log_')
        fpanel = pd.concat([fpanel, lpanel], axis=1).dropna()

        # regress
        ret_rnd = smf.ols('log_rnd ~ log_kappar', data=fpanel).fit()

        if output:
            return fpanel, ret_rnd
        else:
            return -ret_rnd.params['log_kappar']

##
## quick functions
##

def get_moments(params=params_default_fname, target=target_default_fname, drop=True, **kwargs):
    econ = Economy(params=params)
    eq, ts = econ.equilibrium(**kwargs)
    panel = econ.generate_panel(eq)
    moments = econ.calc_moments(panel, frame=True, infer=True, target=target)
    if drop:
        return moments.loc[mmt_names]
    else:
        return moments

def get_jacobian(params=params_default_fname, step=0.05, tdel='prop', tjac='prop'):
    def meval(par):
        econ = Economy(params=par)
        eq, ts = econ.equilibrium()
        panel = econ.generate_panel(eq)
        mmt_sim = econ.calc_moments(panel, infer=True)
        return mmt_sim

    par0 = load_toml(params)
    mmt0 = meval(par0)

    if tdel == 'prop':
        pdel = {k: step*v for k, v in par0.items()}
    elif tdel == 'abs':
        pdel = {k: step for k in par0}

    jac = OrderedDict()
    for pname in var_params:
        print(pname)

        par1 = par0.copy()
        par1[pname] = par0[pname] + pdel[pname]
        mmt1 = meval(par1)

        par2 = par0.copy()
        par2[pname] = par0[pname] - pdel[pname]
        mmt2 = meval(par2)

        if tjac == 'prop':
            mmt_frac = 0.5*(pd.Series(mmt1)-pd.Series(mmt2))/pd.Series(mmt0)
        elif tjac == 'abs':
            mmt_frac = (pd.Series(mmt1)-pd.Series(mmt2))/(2*pdel[pname])

        jac[pname] = mmt_frac[mmt_names]

    return pd.DataFrame(jac)
