import sys
import toml
import numpy as np
import pandas as pd
from scipy.interpolate import InterpolatedUnivariateSpline

import matplotlib as mpl
from matplotlib import style
mpl.use('Agg')
style.use('algo/clean.mplstyle')
import matplotlib.pyplot as plt

# command arguments
import argparse
parser = argparse.ArgumentParser(description='Find optimal policies using parallel annealing.')
parser.add_argument('--params', type=str, help='parameter values to use')
parser.add_argument('--policy', type=str, help='policy values to plot')
parser.add_argument('--tag', type=str, default=None, help='tag base to write to')
args = parser.parse_args()

# load optimal policies
params = toml.load(open(args.params))
policy = toml.load(open(args.policy))

# externality / policy options to run
all_zeta = [0.00, 0.01, 0.02]
labels = {
    'lin': 'Linear',
    'linx': 'Linear + Cross',
    'hsvbnd': 'HSV',
    'hsvxbnd': 'HSV + Cross',
    'hsvxbndplus': 'HSV + CrossPlus',
    'hsvxbnddecint': 'HSV + Cross + Time',
}
all_pols = list(labels)

def plot_absolute(df, ax):
    df = np.log(df)
    df = df.rename(labels, axis=1)
    df = df.rename_axis('Policy Type', axis=1)
    df = df.rename_axis('Externality ($\\zeta$)')
    df.plot(ax=ax, marker='o')
    ax.set_ylabel('Log Welfare')
    ax.legend(loc='upper left', bbox_to_anchor=(1, 1))

def plot_relative(df, ax):
    df = df[all_pols].div(df['opt'], axis=0)
    df = df.rename(labels, axis=1)
    df = df.rename_axis('Policy Type', axis=1)
    df = df.rename_axis('Externality ($\\zeta$)')
    df.plot(ax=ax, marker='o')
    ax.hlines(1.0, *ax.get_xlim(), linestyle='--', linewidth=1)
    ax.set_ylim(0.0, 1.0)
    ax.set_ylabel('Relative Welfare')
    ax.legend(loc='upper left', bbox_to_anchor=(1, 1))

# data store
data = []
for ptype in ['opt'] + all_pols:
    if ptype in policy:
        pinfo = policy[ptype]

        for zeta in reversed(all_zeta):
            if zeta is None:
                zeta = params['zeta']
                ztag = 'baseline'
            else:
                ztag = f'zeta{100*zeta:02.0f}'

            mtag = f'{ztag}_monopoly'
            if mtag in pinfo:
                welf_monopoly = pinfo[mtag]['welfare']
                data.append(('monopoly', ptype, zeta, welf_monopoly))

            ctag = f'{ztag}_control'
            if ctag in pinfo:
                welf_control = pinfo[ctag]['welfare']
                data.append(('control', ptype, zeta, welf_control))

# final output
df_welf = pd.DataFrame(data, columns=['prod', 'ptype', 'zeta', 'welfare'])
df_welf = df_welf.set_index(['prod', 'ptype','zeta']).unstack(level=[0, 1])['welfare']

if args.tag is not None:
    # graph absolute results
    fig, ax = plt.subplots(figsize=(4, 4))
    plot_absolute(df_welf['monopoly'], ax)
    fig.savefig(f'images/welfare_absolute_{args.tag}_monopoly.png', bbox_inches='tight')

    fig, ax = plt.subplots(figsize=(4, 4))
    plot_absolute(df_welf['control'], ax)
    fig.savefig(f'images/welfare_absolute_{args.tag}_control.png', bbox_inches='tight')

    # graph relative results
    fig, ax = plt.subplots(figsize=(4, 4))
    plot_relative(df_welf['monopoly'], ax)
    fig.savefig(f'images/welfare_relative_{args.tag}_monopoly.png', bbox_inches='tight')

    fig, ax = plt.subplots(figsize=(4, 4))
    plot_relative(df_welf['control'], ax)
    fig.savefig(f'images/welfare_relative_{args.tag}_control.png', bbox_inches='tight')
